package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.box.Box
import com.ionspin.kotlin.crypto.box.BoxCorruptedOrTamperedDataException
import com.ionspin.kotlin.crypto.box.crypto_box_NONCEBYTES
import com.ionspin.kotlin.crypto.scalarmult.ScalarMultiplication
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.bipack.BipackDecoder
import net.sergeych.bipack.BipackEncoder
import net.sergeych.crypto2.Asymmetric.Message
import net.sergeych.crypto2.Asymmetric.PublicKey
import net.sergeych.crypto2.Asymmetric.SecretKey

/**
 * Public-key encryption implementation. Generally should be libsodium-compatible.
 *
 * ## How to
 *
 * - [Asymmetric.generateKeys] create a key pair. Keys are serializable.
 * - [SecretKey] provides authenticated encryption for a [PublicKey] receiver.
 * - [PublicKey] provides decryption and anonymous encryption.
 * - [Message] is a serializable container with encrypted message and all necessary data to decrypt it.
 *
 * __Algorithms:__
 *
 * - Key exchange: X25519.
 * - Encryption: XSalsa20
 * - Authentication: Poly1305
 */
object Asymmetric {

    /**
     * Encrypted message holder.
     *
     * Do not instantiate it directly, use [PublicKey.encryptMessage], [PublicKey.encryptAnonymousMessage], etc.
     * instead. Also [SecretKey.decrypt] can be used to decrypt it same as [decrypt] or [decryptWithSenderKey].
     *
     * To successfully decrypt the message, it is necessary to know a sender public key, and non-secret nonce.
     * This class carries all this information; serialize and pass it to the recipient.
     */
    @Serializable
    class Message(
        private val nonce: UByteArray,
        private val encryptedMessage: UByteArray,
        val senderPublicKey: PublicKey,
    ) {
        /**
         * Decrypt the message, same as [SecretKey.decrypt]
         */
        fun decrypt(recipientKey: SecretKey): UByteArray {
            return decryptWithSenderKey(senderPublicKey, recipientKey)
        }

        /**
         * Decrypt a message which is not include sender's public key (which should somehow be
         * known to the recipient). Use it if [senderPublicKey] is null.
         */
        fun decryptWithSenderKey(senderKey: PublicKey, recipientKey: SecretKey): UByteArray {
            return try {
                WithFill.decode(
                    Box.openEasy(encryptedMessage, nonce, senderKey.keyBytes, recipientKey.keyBytes)
                )
            } catch (_: BoxCorruptedOrTamperedDataException) {
                throw DecryptionFailedException()
            }
        }

        val encoded: UByteArray by lazy { BipackEncoder.encode(this).toUByteArray() }
    }

    /**
     * Encrypt the [plainData] using [from] sender for [recipient] public key. Note that to decrypt it
     * the [SecretKey] that corresponds to the [recipient] public key is needed, Sender can't decrypt the message!
     *
     * The authenticated encryption is used, is the message _is successfully decrypted_, it also means that
     * it was signed by the sender, whose public key is known at the decryption time.
     *
     * When it is important not to provide senders' key, use [PublicKey.encryptAnonymousMessage].
     *
     * @param from the senders' secret key.
     * @param recipient the recipients' public key.
     * @param plainData data to encrypt
     */
    private fun createMessage(
        from: SecretKey, recipient: PublicKey, plainData: UByteArray,
        nonce: UByteArray = randomNonce(),
    ): Message {
        return Message(
            nonce,
            Box.easy(plainData, nonce, recipient.keyBytes, from.keyBytes),
            from.publicKey
        )
    }

    /**
     * The generated key pair. See [generateKeys]
     */
    data class KeyPair(val secretKey: SecretKey, val publicKey: PublicKey)

    /**
     * Generate a new random pair of public and secret keys.
     */
    fun generateKeys(): KeyPair {
        val p = Box.keypair()
        val pk = PublicKey(p.publicKey)
        return KeyPair(SecretKey(p.secretKey, pk), pk)
    }

    private fun randomNonce(): UByteArray = randomUBytes(crypto_box_NONCEBYTES)

    fun new(): SecretKey = generateKeys().secretKey

    @Suppress("unused")
    val nonceBytesLength = crypto_box_NONCEBYTES

    /**
     * The public key: [encryptMessage] so only a secret key owner can read it. Allows
     * anonymous [encryptAnonymousMessage] and signed [encryptMessage] encryption.
     *
     * Anonymous encryption is very slow in comparison.
     */
    @Serializable
    class PublicKey(override val keyBytes: UByteArray) : BinaryKeyBase() {

        override val magick: KeysMagickNumber = KeysMagickNumber.defaultAssymmetric

        /**
         * Create an anonymous message that could be decrypted only with the [SecretKey] that corresponds this.
         * Anonymous message uses one-time secret key, the public part of which is included into the
         * [Message], so the sender could not be identified.
         *
         * __Anonymous encryption is much slower__ as it generates new keys every time, use [encryptMessage]
         * when possible
         *
         * The authentication is used despite the anonymity, and the fact of the successful decryption
         * proves that the message was not altered after creation.
         */
        fun encryptAnonymousMessage(plainData: UByteArray, randomFill: IntRange? = null): Message =
            encryptMessage(plainData,randomFill=randomFill)

        /**
         * Anonymous encryption, see [encryptAnonymousMessage], to binary data. Sender could not be identified.
         */
        @Suppress("unused")
        fun encrypt(plainData: UByteArray, randomFill: IntRange? = null): UByteArray =
            encryptMessage(plainData, randomFill = randomFill).encoded

        /**
         * Universal public-key encryption. Note that message authenticity is guaranteed if the decryption is successful
         * whether [senderKey] is provider, the latter only allow to positively identify the sender.
         *
         * @param plainData data to encrypt
         * @param nonce allows specifying exact nonce, default to random (safe)
         * @param senderKey key to authenticate sending party. It is safe and much faster to specify it,
         *      otherwise an anonymous key will be created for each encryption, also safe and anonymous, but slow.
         */
        fun encryptMessage(
            plainData: UByteArray,
            nonce: UByteArray = randomNonce(),
            senderKey: SecretKey = new(),
            randomFill: IntRange? = null,
        ) = createMessage(senderKey, this, WithFill.encode(plainData, randomFill), nonce)

        /**
         * Encrypt message using the specified secret key as sender authentication. Recipient, the party having
         * [SecretKey] corresponding to this one, will be able to decrypt the message and be sure that [senderKey]
         * was the author and the message was not altered.
         */
        fun encryptMessage(plainData: UByteArray,
                           senderKey: SecretKey,
                           randomFill: IntRange? = null): Message =
            createMessage(senderKey, this, WithFill.encode(plainData, randomFill))

    }

    /**
     * The secret key
     */
    @Serializable
    class SecretKey(
        override val keyBytes: UByteArray,
        @Transient
        val _cachedPublicKey: PublicKey? = null,
    ) : DecryptingKey, BinaryKeyBase() {

        /**
         * Decrypt with authentication checks the message which must have [Message.senderPublicKey] set.
         * Use [decryptWithSenderKey] otherwise. Note that the authenticated encryption is always use, even if
         * the [PublicKey.encryptAnonymousMessage] was used to create a message, if it is successfully decrypted,
         * it is guaranteed that the message was not altered after creation.
         *
         * @throws DecryptionFailedException If the message is tampered (changed after creation) or was not intended for us,
         */
        fun decrypt(message: Message): UByteArray = message.decrypt(this)


        /**
         * Decrypt using [senderPublicKey] as a sender key (overriding the [Message.senderPublicKey] if set).
         * See [decrypt] for more.
         */
        fun decryptWithSenderKey(message: Message, senderPublicKey: PublicKey): UByteArray =
            message.decryptWithSenderKey(senderPublicKey, this)

        @Transient
        private var cachedPublicKey: PublicKey? = _cachedPublicKey

        /**
         * The corresponding public key
         */
        val publicKey: PublicKey by lazy {
            PublicKey(ScalarMultiplication.scalarMultiplicationBase(keyBytes))
                .also { cachedPublicKey = it }
        }

        /**
         * Nonce-based decryption is impossible, it is already included in message
         */
        override fun decryptWithNonce(cipherData: UByteArray, nonce: UByteArray): UByteArray = decrypt(cipherData)

        /**
         * Decrypt without a nonce as edwards curve decryption does not need it
         */
        override fun decrypt(cipherData: UByteArray): UByteArray {
            val message: Message = BipackDecoder.decode(cipherData.toByteArray())
            return message.decrypt(this)
        }

        override val magick: KeysMagickNumber = KeysMagickNumber.defaultAssymmetric
        override val id: KeyId by lazy { publicKey.id }

        override val nonceBytesLength: Int
            get() = 0
    }


}

/**
 * Shortcut type: a pair of sender secret key and recipient private key could be used so
 * simplify such interfaces
 */
typealias AsymmetricEncryptionPair = Pair<SecretKey?,PublicKey>
